/*
 * Copyright 2008-2010 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wamblee.glassfish.auth;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.wamblee.glassfish.auth.cache.api.AuthenticationCache;
import org.wamblee.glassfish.auth.cache.impl.NoAuthenticationCache;

/**
 * Utility for creating a cache based on properties. Separated from
 * {@link FlexibleJdbcRealm} for testability.
 * 
 * @author Erik Brakkee
 * 
 */
public class CacheConstructor {

    private static final Logger LOGGER = Logger
        .getLogger(CacheConstructor.class.getName());

    /**
     * Creates the cache using introspection based on the classname and passed
     * the properties to the constructor of the configured cache.
     * 
     * @param aProperties
     *            Properties.
     * @return Configured cache or the default cache in case a problem occured
     *         constructing the cache.
     */
    public AuthenticationCache createCache(Properties aProperties) {
        AuthenticationCache cache = createCacheImpl(aProperties);
        LOGGER.info("Using cache '" + cache + "'");
        return cache;
    }

    private AuthenticationCache createCacheImpl(Properties aProperties) {
        String classname = aProperties
            .getProperty(FlexibleJdbcRealm.PROP_AUTHENTICATION_CACHE_CLASSNAME);
        if (classname == null) {
            return createDefaultCache(aProperties);
        }
        try {
            Class clazz = Class.forName(classname);
            Constructor<?> ctor = clazz.getConstructor(Properties.class);
            Object obj = ctor.newInstance(aProperties);
            if (!(obj instanceof AuthenticationCache)) {
                LOGGER.severe("Class '" + classname +
                    "' is not an instance of '" +
                    AuthenticationCache.class.getName() + "'");
                return createDefaultCache(aProperties);
            }
            return (AuthenticationCache) obj;
        } catch (ClassNotFoundException e) {
            return defaultCacheBecauseOfProblem("Cache class not found '" +
                classname + "'", aProperties, e);
        } catch (NoSuchMethodException e) {
            return defaultCacheBecauseOfProblem(
                "Cache class '" + classname +
                    "' does not have public constructor with java.util.Properties argument",
                aProperties, e);
        } catch (InvocationTargetException e) {
            return defaultCacheBecauseOfProblem(
                "Exception occurred creating cache instance of '" + classname +
                    "'", aProperties, e.getCause());
        } catch (Exception e) {
            return defaultCacheBecauseOfProblem(
                "Exception occurred creating cache instance of '" + classname +
                    "'", aProperties, e);
        }
    }

    private AuthenticationCache defaultCacheBecauseOfProblem(String aMsg,
        Properties aProperties, Throwable aException) {
        LOGGER.log(Level.SEVERE, aMsg + " (properties " + aProperties + ")",
            aException);
        return createDefaultCache(aProperties);
    }

    private AuthenticationCache createDefaultCache(Properties aProperties) {
        LOGGER.info("Using default cache");
        return new NoAuthenticationCache(aProperties);
    }
}
