/*
 * Copyright 2008-2010 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wamblee.glassfish.auth;

/**
 * Parser for the encoding based on a string.
 * 
 * Three types of syntax are supported for the string:
 * <ul>
 * <li>HEX: HEX encoder padding to 32 characters.</li>
 * <li>HEX:nn: HEX encoder padded to nn digits.</li>
 * <li>BASE64: Base 64 encoder.</li>
 * </ul>
 * 
 * @author Erik Brakkee
 */
public class EncoderParser {

    private static final String TEXT = "TEXT";
    private static final String HEX = "HEX";
    private static final String BASE64 = "BASE64";

    private String charset;

    /**
     * Constructs the parser
     * 
     * @param aCharset
     *            Character set.
     */
    public EncoderParser(String aCharset) {
        charset = aCharset;
    }

    public Encoder parse(String aEncodingString) {
        if (aEncodingString == null || HEX.equals(aEncodingString)) {
            return new HexEncoder();
        }
        if (TEXT.equals(aEncodingString)) {
            return new TextEncoder(charset);
        }
        if (BASE64.equals(aEncodingString)) {
            return new Base64Encoder();
        }
        if (aEncodingString.startsWith(HEX + ":")) {
            String remainder = aEncodingString.substring(4);
            int padding = Integer.parseInt(remainder);
            if (padding < 0) {
                throw new IllegalArgumentException(
                    "Negative padding for encoding string '" + aEncodingString +
                        "'");
            }
            return new HexEncoder(padding);
        }
        throw new IllegalArgumentException("Unsupported encoding '" +
            aEncodingString + "', " +
            "valid syntax is 'HEX', 'BASE64', or 'HEX:<digits>");
    }
}
