/*
 * Copyright 2008-2010 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wamblee.glassfish.auth;

import com.sun.appserv.security.AppservRealm;
import java.nio.charset.Charset;
import java.text.MessageFormat;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.wamblee.glassfish.auth.cache.api.AuthenticationCache;

import com.sun.enterprise.security.auth.realm.BadRealmException;
import com.sun.enterprise.security.auth.realm.IASRealm;
import com.sun.enterprise.security.auth.realm.InvalidOperationException;
import com.sun.enterprise.security.auth.realm.NoSuchRealmException;
import com.sun.enterprise.security.auth.realm.NoSuchUserException;

/**
 * JDBC Realm with persistent storage of users and groups in a database. The
 * realm is configured with two database queries:
 * <ul>
 * <li>A query to retrieve the possibly encoded password of the username.</li>
 * <li>A query to retrieve the groups for a user based on username.</li>
 * </ul>
 * 
 * It supports plain encoding of the password and MD5 HEX encoding.
 * 
 * @author Erik Brakkee
 */
// TODO Should add a parameter for defining the group cache.
public class FlexibleJdbcRealm extends AppservRealm {

    private static final Logger LOGGER = Logger
        .getLogger(FlexibleJdbcRealm.class.getName());

    /**
     * Property name for the realm name.
     */
    private static final String PROP_JAAS_CONTEXT = "jaas.context";

    /**
     * Property name for the datasource JNDI name.
     */
    private static final String PROP_DATASOURCE_JNDI = "datasource.jndi";

    /**
     * Property for the query for retrieving the encoded password. This should
     * be a query in prepared statement syntax with one parameter for the user
     * name.
     */
    private static final String PROP_SQL_PASSWORD = "sql.password";

    /**
     * Property for the query for retrieving the groups for a given user name.
     * This should be a query in prepared statement syntax with one parameter
     * for the user name.
     */
    private static final String PROP_SQL_GROUPS = "sql.groups";

    /**
     * Property for the query for retrieving the seed for a given user name.
     * This should be a query in prepared statement syntax with one parameter
     * for the user name.
     */
    private static final String PROP_SQL_SEED = "sql.seed";

    /**
     * Property that describes the final string that will be encoded. This
     * string is a string in {@link MessageFormat} format with the first
     * parameter the password and the second parameter the seed.
     */
    private static final String PROP_SEED_FORMAT = "password.seed.format";

    /**
     * Default seed format string.
     */
    private static final String DEFAULT_SEED_FORMAT = "{0}{1}";

    /**
     * Digest method used for the password.
     */
    private static final String PROP_DIGEST = "password.digest";

    /**
     * Encoding method used for the digest, see {@link EncoderParser}
     */
    private static final String PROP_ENCODING = "password.encoding";

    /**
     * Charset name to use. When null, the default charset is used.
     */
    private static final String PROP_CHARSET = "charset";

    /**
     * Optional property for the definition of a comma-separated list of groups
     * to which each authenticated user is added. This property is implemented
     * by the base realm.
     */
    private static final String PROP_ADDITIONAL_GROUPS = "assign-groups";

    /**
     * Optional property for the classname of the optional caching class to use.
     */
    static final String PROP_AUTHENTICATION_CACHE_CLASSNAME = "cache.classname";

    private String jaasContext;

    private RealmWorker worker;

    public FlexibleJdbcRealm() {
        // Empty.
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * com.sun.enterprise.security.auth.realm.Realm#init(java.util.Properties)
     */

    protected void init(Properties aProperties) throws BadRealmException,
        NoSuchRealmException {
        super.init(aProperties);
        LOGGER.log(Level.INFO, "Initializing flexible JDBC realm: {0}", aProperties.toString());
        jaasContext = getProp(aProperties, PROP_JAAS_CONTEXT);
        this.setProperty(IASRealm.JAAS_CONTEXT_PARAM, jaasContext);

        String digestMethod = getProp(aProperties, PROP_DIGEST);

        String charset = aProperties.getProperty(PROP_CHARSET);
        if (charset == null) {
            charset = Charset.defaultCharset().name();
        }

        String encodingString = aProperties.getProperty(PROP_ENCODING);
        Encoder encoder = new EncoderParser(charset).parse(encodingString);

        String seedFormat = aProperties.getProperty(PROP_SEED_FORMAT);
        if (seedFormat == null || seedFormat.length() == 0) {
            seedFormat = DEFAULT_SEED_FORMAT;
        }

        String datasource = getProp(aProperties, PROP_DATASOURCE_JNDI);
        String sqlPasswordQuery = getProp(aProperties, PROP_SQL_PASSWORD);
        String sqlGroupsQuery = getProp(aProperties, PROP_SQL_GROUPS);
        String sqlSeedQuery = aProperties.getProperty(PROP_SQL_SEED);
        if (sqlSeedQuery != null && sqlSeedQuery.length() == 0) {
            sqlSeedQuery = null;
        }

        final String[] additionalGroups = extractAdditionalGroups(aProperties);

        AuthenticationRepository jdbcRepo = new JdbcAuthenticationRepository(
            datasource, sqlPasswordQuery, sqlGroupsQuery, sqlSeedQuery);
        AuthenticationRepository authRepo = new CachedAuthenticationRepository(
            createCache(aProperties), jdbcRepo);

        worker = new RealmWorker(charset, seedFormat, digestMethod, encoder,
            additionalGroups, authRepo);

        LOGGER.info("Flexible JDBC realm '" + jaasContext + "' initialized");
    }

    static AuthenticationCache createCache(Properties aProperties) {
        return new CacheConstructor().createCache(aProperties);
    }

    private String[] extractAdditionalGroups(Properties aProperties) {
        String additionalGroupsString = aProperties
            .getProperty(PROP_ADDITIONAL_GROUPS);
        String[] additionalGroups = new String[0];
        if (additionalGroupsString != null) {
            additionalGroups = additionalGroupsString.split(",");
        }
        return additionalGroups;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.sun.enterprise.security.auth.realm.Realm#toString()
     */
    @Override
    public String toString() {
        return "FlexibleJdbcRealm(jaasContext = " + jaasContext +
            ", worker = " + worker + ")";
    }

    private String getProp(Properties aProps, String aKey)
        throws BadRealmException {
        String value = aProps.getProperty(aKey);
        if (value == null) {
            String msg = "Missing property '" + aKey +
                "' for FlexibleJdbcRealm";
            throw new BadRealmException(msg);
        }
        if (LOGGER.isLoggable(Level.FINEST)) {
            LOGGER
                .finest("Config property read " + aKey + " = '" + value + "'");
        }
        return value;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * com.sun.enterprise.security.auth.realm.Realm#getGroupNames(java.lang.
     * String)
     */
    @Override
    public Enumeration getGroupNames(String aUsername)
        throws InvalidOperationException, NoSuchUserException {

        Vector<String> groupNames = worker.getGroupNames(aUsername);
        if (LOGGER.isLoggable(Level.FINEST)) {
            LOGGER.finest("User " + aUsername + " is in groups " +
                groupNames.toString());
        }
        return groupNames.elements();
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.sun.enterprise.security.auth.realm.Realm#getAuthType()
     */
    @Override
    public String getAuthType() {
        return "Flexible JDBC authentication against any type of database and schema";
    }

    public String[] authenticateUser(String aUsername, String aPassword) {

        String[] groups = worker.authenticateUser(aUsername, aPassword);
        if (LOGGER.isLoggable(Level.FINEST)) {
            if (groups != null) {
                LOGGER.finest("Authenticate user '" + aUsername +
                    "' is in groups " + Arrays.asList(groups));
            } else {
                LOGGER.finest("Authenticate use '" + aUsername + "' failed");
            }
        }
        return groups;
    }

}
