/*
 * Copyright 2008-2010 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wamblee.glassfish.auth;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.naming.InitialContext;
import javax.security.auth.login.LoginException;
import javax.sql.DataSource;

/**
 * Implementation of the authentiction repository through a relational database
 * with database queries.
 * 
 * @author Erik Brakkee
 */
public class JdbcAuthenticationRepository implements AuthenticationRepository {

    private static final Logger LOGGER = Logger
        .getLogger(JdbcAuthenticationRepository.class.getName());

    private String datasource;

    private String sqlPasswordQuery;

    private String sqlGroupsQuery;

    private String sqlSeedQuery;

    public JdbcAuthenticationRepository(String aDatasource,
        String aSqlPasswordQuery, String aSqlGroupsQuery, String aSqlSeedQuery) {
        datasource = aDatasource;
        sqlPasswordQuery = aSqlPasswordQuery;
        sqlGroupsQuery = aSqlGroupsQuery;
        sqlSeedQuery = aSqlSeedQuery;
    }

    @Override
    public List<String> findGroups(String aUsername) {      
        List<String> groups = executeQuery(sqlGroupsQuery, aUsername);
        if (groups == null) {
            groups = new ArrayList<String>();
        }
        return groups;
    }

    @Override
    public String findPassword(String aUsername) {
        List<String> passwords = executeQuery(sqlPasswordQuery, aUsername);
        if (passwords == null || passwords.size() != 1) {
            if (LOGGER.isLoggable(Level.FINEST)) {
                LOGGER.finest("No password found for '" + aUsername + "'");
            }
            return null;
        }
        return passwords.get(0);
    }

    @Override
    public String findSeed(String aUsername) {
        if (LOGGER.isLoggable(Level.FINEST)) {
            LOGGER.finest("Finding seed for user '" + aUsername + "'");
        }
        if (sqlSeedQuery == null) {
            if (LOGGER.isLoggable(Level.FINEST)) {
                LOGGER.finest("No seed query defined so no seed");
            }
            return "";
        }
        List<String> seeds = executeQuery(sqlSeedQuery, aUsername);
        if (seeds == null || seeds.size() != 1) {
            if (LOGGER.isLoggable(Level.FINEST)) {
                LOGGER.finest("Multiple seeds returned for user '" + aUsername +
                    "' seeds are: " + seeds);
            }
            return "";
        }
        return seeds.get(0);
    }

    private List<String> executeQuery(String aQuery, String aUsername) {
        if (LOGGER.isLoggable(Level.FINEST)) {
            LOGGER.finest("Executing query '" + aQuery + "' for user '" +
                aUsername + "'");
        }
        Connection connection = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        try {
            connection = getConnection();
            statement = connection.prepareStatement(aQuery);
            statement.setString(1, aUsername);
            rs = statement.executeQuery();
            final List<String> result = new ArrayList<String>();
            while (rs.next()) {
                result.add(rs.getString(1));
            }
            return result;
        } catch (Exception ex) {
            LOGGER.log(Level.SEVERE, "Error in query '" + aQuery +
                "' for user " + aUsername + "Cannot load group", ex);
            return null;
        } finally {
            close(connection, statement, rs);
        }
    }

    /**
     * Return a connection from the properties configured
     * 
     * @return a connection
     */
    private Connection getConnection() throws LoginException {

        try {
            InitialContext ctx = new InitialContext();
            final DataSource dataSource = (DataSource) ctx.lookup(datasource);
            Connection connection = dataSource.getConnection();
            return connection;
        } catch (Exception ex) {
            String msg = "FlexibleJdbcRealm Cannot connect, jndi = " +
                datasource;
            LoginException loginEx = new LoginException(msg);
            loginEx.initCause(ex);
            throw loginEx;
        }
    }

    private void close(Connection aConnection, PreparedStatement aStatement,
        ResultSet aResultSet) {
        if (aResultSet != null) {
            try {
                aResultSet.close();
            } catch (Exception e) {
                LOGGER.log(Level.WARNING, "Error closing result set", e);
            }
        }

        if (aStatement != null) {
            try {
                aStatement.close();
            } catch (Exception e) {
                LOGGER.log(Level.WARNING, "Error closing statement", e);
            }
        }

        if (aConnection != null) {
            try {
                aConnection.close();
            } catch (Exception e) {
                LOGGER.log(Level.WARNING, "Error closing result set", e);
            }
        }
    }

    @Override
    public String toString() {
        return "JdbcAuthenticationRepository(" + "dataSource = " + datasource +
            ", sqlPasswordQuery = " + sqlPasswordQuery + ", sqlGroupsQuery = " +
            sqlGroupsQuery + ", sqlSeedQuery = " + sqlSeedQuery + ")";
    }
}
