/*
 * Copyright 2008-2010 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wamblee.glassfish.auth;

import java.io.UnsupportedEncodingException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.MessageFormat;
import java.util.Arrays;
import java.util.List;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * POJO based realm implementation that gets all its dependencies wired in. This
 * is good for testability.
 * 
 * @author Erik Brakkee
 */
public class RealmWorker {

    private static final Logger LOGGER = Logger.getLogger(RealmWorker.class
        .getName());

    /**
     * Reserved string for plain encoding.
     */
    private static final String PLAIN_ENCODING = "PLAIN";

    private String charset;

    private String seedFormat;

    private String digestMethod;

    private Encoder encoder;

    private String[] additionalGroups;

    private AuthenticationRepository authRepo;

    /**
     * Constructs the realmworker.
     * 
     * @param aCharset
     *            character set to use.
     * @param aSeedFormat
     *            Seedformat
     * @param aDigestMethod
     *            Digest method for password.
     * @param aEncoder
     *            Encoder to use of the password.
     * @param aAdditionalGroups
     *            Additional groups every user should belong to.
     * @param aAuthRepo
     */
    public RealmWorker(String aCharset, String aSeedFormat,
        String aDigestMethod, Encoder aEncoder, String[] aAdditionalGroups,
        AuthenticationRepository aAuthRepo) {
        charset = aCharset;
        seedFormat = aSeedFormat;
        digestMethod = aDigestMethod;
        encoder = aEncoder;
        additionalGroups = aAdditionalGroups;
        authRepo = aAuthRepo;
    }

    /**
     * Gets the group names for a given user, also adding the additional groups
     * to the list.
     * 
     * @param aUsername
     *            Username. Note: assumed is that a valid user is always passed.
     * @return List of groups the user belongs to. Always non-null.
     */
    public Vector<String> getGroupNames(String aUsername) {
        String[] groups = findGroups(aUsername);
        return new Vector<String>(Arrays.asList(groups));
    }

    private String[] findGroups(String aUsername) {
        List<String> groups = authRepo.findGroups(aUsername);
        groups.addAll(Arrays.asList(additionalGroups));
        return groups.toArray(new String[0]);
    }

    /**
     * Authenticate the user.
     * 
     * @param aUsername
     *            Username.
     * @param aPassword
     *            Password.
     * @return List of groups the user belongs to or null if the user is not
     *         authenticated.
     */
    public String[] authenticateUser(String aUsername, String aPassword) {
        String[] groups = null;
        if (isUserValid(aUsername, aPassword)) {
            groups = findGroups(aUsername);
        }
        return groups;
    }

    /**
     * Test if a user is valid
     * 
     * @param aUsername
     *            user's identifier
     * @param aPassword
     *            user's password
     * @return true if valid
     */
    private boolean isUserValid(final String aUsername, final String aPassword) {
        String retrievedPassword = authRepo.findPassword(aUsername);
        if (retrievedPassword == null) {
            return false;
        }
        String encodedPassword;

        // Apply seeding to the password.
        String seed = authRepo.findSeed(aUsername);
        String passwordWithSeed = MessageFormat.format(seedFormat, aPassword,
            seed);

        if (digestMethod.equals(PLAIN_ENCODING)) {
            encodedPassword = passwordWithSeed;
        } else {
            encodedPassword = encodePassword(passwordWithSeed, digestMethod);
        }

        if (LOGGER.isLoggable(Level.FINEST)) {
            LOGGER.finest("Encoded password for '" + aUsername + "' is '" +
                encodedPassword + "'");
        }
        return encodedPassword.equals(retrievedPassword);
    }

    /**
     * Encodes the password using the given encoding.
     * 
     * @param aPassword
     *            Password.
     * @param aEncoding
     *            Encoding to use.
     * @return Encoded password.
     */
    private String encodePassword(String aPassword, String aEncoding) {
        try {
            MessageDigest md = MessageDigest.getInstance(aEncoding);
            byte[] digested = md.digest(aPassword.getBytes(charset));
            return encoder.encode(digested);
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException("Could not get encoding '" + aEncoding +
                "' for encoding password", e);
        } catch (UnsupportedEncodingException e) { 
            throw new RuntimeException("Unsupported charset encoding '" + charset + "'"); 
        }

    }

    /*
     * (non-Javadoc)
     * 
     * @see com.sun.enterprise.security.auth.realm.Realm#toString()
     */
    @Override
    public String toString() {
        return "FlexibleJdbcRealm(digestMethod = " + digestMethod +
            ", encoder = " + encoder.toString() + ", seedFormat = " +
            seedFormat + ", authRepo = " + authRepo + ")";
    }
}
